"""Numerical methods for mathematical functions neede for the program"""

#This method finds the factorial of a given number(num).
#The method calls for an integer input num
#The method returns the factorial(represented as fact in the method)
def factorial(num):
    """Finds the factorial of the input integer.
    
    :arg num: an integer
    """
    #If the number provided is zero then the factorial is 1
    if num == 0:
	fact = 1
    #Otherwise set fact to 1 and begin finding the factorial r is
    #used to find each num-n for n=0 to n=num each value in r is
    #then used to compute the factorial
    else:
	fact = 1
	r = range(1,num+1)
	for i in r:
	    fact *= i
    return fact

#This method finds a binomial coefficient that is it calculates n
#choose r The method calls for two integer values n and r and the
#method factorial The method returns the binomial factorial as result
def binomial_coefficient(n,r):
    """Finds the binomial coefficient, n choose r, for a given set of
    integers.
    
    :arg n: an integer
    :arg r: an integer
    """
    #If r is less than zero then the binomial coefficient is zero
    if r < 0:
	result = 0
    #Otherwise use the factorial method to calculate the binomial
    #coefficient n!/r!/(n-r)!
    else:
	result = factorial(n) / factorial(r) / factorial(n-r)
    return result



"""This module handles writing to the terminal or a log file with support
for coloring for warnings, errors, etc."""
from termcolor import cprint
verbosity = None
"""The verbosity level of messages being printed by the module."""
quiet = None
"""When in quiet mode, no text is ever printed to terminal unless its
verbosity level is explicitly specified.
"""
cenum = {
    "cerrs": 0,
    "cwarn": 1,
    "cinfo": 2,
    "cgens": 3,
    "cstds": 4,
    "cokay": 5
}
"""Dict of the various colors available for coloring specific
lines in the arb().
"""
icols = ["red", "yellow", "cyan", "blue", "white", "green"]

def arb(text, cols, split):
    """Prints a line of text in arbitrary colors specified by the numeric
    values contained in msg.cenum dictionary.
    """
    stext = text if text[-1] != split else text[0:-1]
    words = stext.split(split)
    for i, word in enumerate(words):
        col = icols[cols[i]]
        cprint(word, col, end="")
        if i < len(words)-1:
            cprint(split, end="")
        else:
            cprint(split)

def set_verbosity(level):
    """Sets the modules message verbosity level for *all* messages printed.

    :arg level: a positive integer (>0); higher levels including more detail.
    """
    global verbosity
    verbosity = level

def set_quiet(is_quiet):
    """Sets whether the messaging system is running in quiet mode. Quiet mode
    only prints messages with explicit verbosity specified. If verbosity==None
    and quiet==True, any message with level >= 0 is printed.
    """
    global quiet
    quiet = is_quiet

def will_print(level=1):
    """Returns True if the current global status of messaging would print a
    message using any of the printing functions in this module.
    """
    if level == 1:
        #We only affect printability using the quiet setting.
        return quiet is None or quiet == False
    else:
        return ((isinstance(verbosity, int) and level <= verbosity) or
                (isinstance(verbosity, bool) and verbosity == True))
    
def warn(msg, level=0, prefix=True):
    """Prints the specified message as a warning; prepends "WARNING" to
    the message, so that can be left off.
    """
    if will_print(level):
        cprint(("WARNING: " if prefix else "") + msg, "yellow")

def err(msg, level=-1, prefix=True):
    """Prints the specified message as an error; prepends "ERROR" to
    the message, so that can be left off.
    """
    if will_print(level) or verbosity is None:
        cprint(("ERROR: " if prefix else "") + msg, "red")

def info(msg, level=1):
    """Prints the specified message as information."""
    if will_print(level):
        cprint(msg, "cyan")

def okay(msg, level=1):
    """Prints the specified message as textual progress update."""
    if will_print(level):
        cprint(msg, "green")

def gen(msg, level=1):
    """Prints the message as generic output to terminal."""
    if will_print(level):
        cprint(msg, "blue")

def blank(n=1, level=2):
    """Prints a blank line to the terminal."""
    if will_print(level):
        for i in range(n):
            print("")

def std(msg, level=1):
    """Prints using the standard print() function."""
    if will_print(level):
        print(msg)



"""Methods for reading and writing enumeration and polya-counting results."""
def read_group(fname):
    """Reads the symmetry group in from the 'rot_perms' styled group
    output by enum.x.

    :arg fname: path to the file to read the group from.
    """
    i=0
    groupi = []
    with open(fname) as f:
        for line in f:
            if i > 5:
                if ('Perm #:') in line:
                    groupi.append(map(int, line.split()[4::]))
                else:
                    groupi[-1] += map(int, line.split())
            i += 1
    from numpy import array
    return(map(list, array(groupi)-1))

def read_lattice(filename='lattice.in', verbose=False):
    """Reads the lattice.in file; returns a dictionary with the following fields:
      "sizes": the range of cell sizes,
      "lat_vecs": lattice vectors of the parent cell,
      "nspecies": the number of atomic species in the enumeration,
      "basis_vecs": basis vectors of the parent cell,
      "is_crestricted": logical that indicates if the concentrations will be restricted,
      "arrows": logical that indicates if arrows are present,
      "concs": array of the concentrations in format [1,3].
    """
    from msg import info
    with open(filename,'r') as f:
        lines = [l for l in f.readlines() if l.strip()[0] != '#']
        
    sizes = [int(el) for el in lines[0].split(' ')]
    bulk = lines[1].strip()[0].lower() == 'b'
    lat_vecs = [map(float, l.split()) for l in lines[2:5]]
    nspecies = int(lines[5])
    nbas = int(lines[6])
    bas_vecs = [map(float, l.split()) for l in lines[7:7+nbas]]

    arrows = False
    res_conc = False
    if lines[7+nbas].strip() == 'F':
        info('No concentration restrictions given.')
        concs = []
    else:
        res_conc = True
        info('Concentrations are being restrictied.')
        arrows = lines[7+nbas+1].strip()
        if arrows == 'F':
            arrows = []
            info('No displacement directions are included')
        else:
            arrows = True
        concs = [map(float, l.split()) for l in lines[7+nbas+2:]]

    result = {
        "bulk": bulk,
        "sizes": sizes,
        "lat_vecs": lat_vecs,
        "nspecies": nspecies,
        "basis_vecs": bas_vecs,
        "is_crestricted": res_conc,
        "arrows": arrows,
        "concs": concs
    }
    return result

# This method reads in the enum.in file.
def read_enum(filename="enum.in"):
    """Reads the list of structures and the number of random candidates to enumerate
    from the 'enum.in' styled file. It should contain only integers and have:

       HNF concs num_wanted
    where,

    HNF: the list of 6 independent entries in the HNF matrix;
    concs: list of integer species concentrations;
    num_wanted: number of random structures to draw from the enumerated list.
    """    
    from numpy import loadtxt
    raw = loadtxt(filename, int, ndmin=2)
    systems = []
    for i in range(len(raw)):
        HNF = raw[i,0:6]
        sys_conc = raw[i,6:-1]
        num_wanted = raw[i,-1]
        systems.append((HNF, sys_conc, num_wanted))
    return systems

def write_enum(params, outfile="enum.out"):
    """Writes a 'struct_enum.out' style preamble to the specified output file
    using the enumeration parameters gleaned from a 'lattice.in' styled file.

    :arg params: values returned from method:read_lattice().
    :arg outfile: path to desired output file.
    """
    from datetime import datetime
    lines = []
    lines.append("Random structure enumeration: %s" % (str(datetime.now())))
    lines.append("bulk" if params["bulk"] else "surface")
    lines.extend([" {0:.7f}       {1:.7f}       {2:.7f}".format(*l) for l in params["lat_vecs"]])
    lines.append("    {0:d}".format(len(params["basis_vecs"])))
    lines.extend(["  {0:.7f}       {1:.7f}       {2:.7f}".format(*l) for l in params["basis_vecs"]])
    lines.append(" {0:d}-nary case".format(params["nspecies"]))
    lines.append("   " + '   '.join(map(str, params["sizes"])))
    lines.append(" 1e-7")
    lines.append("Concentration check:")
    lines.append("    T" if params["is_crestricted"] else "    F")
    lines.extend(["full list of labelings (including incomplete labelings) is used ",
                  "Equivalency list: 1 #Not used in the random enumeration algorithm ",
                  "start   #tot      HNF     Hdegn   labdegn   Totdegn   #size idx    "
                  "pg    SNF             HNF                 Left "
                  "transform                          labeling                 "
                  "directions",
                  ""])
        
    with open(outfile, 'w') as f:
        f.write('\n'.join(lines))

def write_struct_enum(params):
    """Writes a 'struct_enum.in' file for the executable enum.x.

    :arg params: values returned from method:read_lattice().
    """

    with open("struct_enum.in","w+") as struct_file:
        struct_file.write("System \n")

        if params["bulk"] == True:
            struct_file.write("bulk \n")
        else:
            struct_file.write("surface \n")

        latt = ""
        i = 0
        for vec in params["lat_vecs"]:
            for point in vec:
                latt += str(point) + " "
            if i < 2:
                latt += "\n"
            i += 1

        struct_file.write(latt + " \n")

        struct_file.write(str(params["nspecies"]) + " \n")

        occupancy = ""
        for i in range(params["nspecies"]):
            if i < params["nspecies"]-1:
                occupancy += str(i) +"/"
            else:
                occupancy += str(i)
    
        basis = ""
        i = 0
        for vec in params["basis_vecs"]:
            for point in vec:
                basis += str(point) + " "
            if i < 2:
                basis += " "+ occupancy + "\n"
            i += 1
        
        struct_file.write(str(len(params["basis_vecs"])) + " \n")
        struct_file.write(basis + " \n")

        size_range = ""
        i = 0
        for point in params["sizes"]:
            size_range += str(point) + " "

        struct_file.write(size_range + " \n")
        struct_file.write("0.00001 \n")
        struct_file.write("full \n")

def which(program):
    """Determines where if an executable exists on the users path.
    This code was contributed by Jay at http://stackoverflow.com/a/377028
    :args program: The name, or path for the program
    """
    import os
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None

def create_labeling(config):
    """This routine takes a string of atomic locations as a vector and the
    atomic concentrations and returns a unique labeling.

    :arg config: list of integers describing the arrangement of atoms on
    the lattice.
    """
    label = ''
    arrow =  ''
    if isinstance(config[0],list):
        for i in config:
            label += str(i[1]-1)
            arrow += str(i[0]+1)
    else:
        for i in config:
            label += str(i-1)
            arrow += '0'
    
    return(label,arrow)



"""Methods needed for the enumeration of arrows."""

#import neede modules
from copy import deepcopy

#Find concentration takes a 1D array, col, and returns a 1D array, 
#Concs, containing the number of times each element in col appears 
#in the array.
def _find_concentrations(col):
    """Finds the concentration of a given array of colors.

    :arg col: an integer array of the labeling
    """
    Concs=[]

    tcol=deepcopy(col)
    for i in range(len(tcol)):
        if tcol[i] != 0:
            Concs.append(tcol.count(tcol[i]))
            for j in range(len(tcol)):
                if tcol[j] == tcol[i] and i != j:
                    tcol[j] =0
            tcol[i] =0
    return(Concs)

#Takes an array of colors and arrows and tells us how many different
#colors have arrows on them.
def how_many_arrows(tcol):
    """Determines the number of colors that have arrows on them.

    :arg tcol: a 2D array of the labeling that contains the colors
    and arrows for each site.
    """
    arrows = 0
    
    species = []
    for i in tcol:
        if i[0] >= 0:
            arrows += 1
            if i[1] not in species:
                species.append(i[1])

    aconcs = _find_concentrations(tcol)
    return(arrows,len(species),aconcs)

#An algorithm for sorting the col array so that colors without arraws
#appear first in the list by concentration followed by the colors with
#arraws that are also sorted by concentration.
def _col_sort(col_list):
    """Sorts the labeling so that colors with arrows appear last in the
    list and so that the arrowed and non-arrowed colors are sorted
    from lowest to highest concentration.

    :arg col_list: a 2D integer array of the full labeling of the
    system
    """

    col1 = []
    col2 = []
    colt=[]

    # seperate the arrays into colors with arrrows and colors without
    for i in col_list:
        if i[0] < 0:
            col1.append(i)
        elif i[0] >= 0:
            col2.append(i)

    # sort each array by concentration
    col1 = sorted(col1, key = col1.count, reverse=True)
    col2 = sorted(col2, key = col2.count, reverse=True)

    # put them back together again
    for i in col1:
        colt.append(i)
    for i in col2:
        colt.append(i)
    
    return(colt)

#this method takes a configuration and finds the unique ways of
#placing the arrows on that configuration.
#col is the initial configuration.
#agroup is the group operations with their effects on the arrows.
#dim is the number of arrow directions that are possible for this system.
def add_arrows(col,agroup,dim):
    """Finds the unique arrangements of arrows for a given configuration.

    :arg col: a 2D integer array of the initial labeling
    :arg agroup: the stabilizers for the colors only with the arrow
    permutations
    :arg dim: the number of directions the arrows can point
    """

    with open("add_arrows_group.out","w+") as af:
        for i in agroup:
            af.write(str(i) + "\n")
    
    #survivors is the array that contains the end result of the permutations
    survivor = []

    #Find out how many arrows there are in the col array
    (narrows,arrow_types,conc_w_arrows) = how_many_arrows(col)
    #largest_arrow is largest value any arrow could have.
    largest_arrow = [dim-1]*narrows
    #this is an array for storing the unique arrangements.
    arsurvivors = []

    #this loop runs over every possible configuration of arrows
    #staring from the one with the smallest hash number and ending
    #with the one with the largest hash number to see if they are
    #unique. Each hash will be stored in orighash (original hash) then
    #compared to the permuted arrows hash (permhash) at the end.
    for orighash in range(_ahash(largest_arrow,dim)+1):
        arrow_config = _ainvhash(orighash,narrows,dim)
        unique = True
        temp_coloring_with_arroms = [0]*len(col)
        #we need to cycle through each of the operations in the
        #permutation group that got passed in to see if they will turn
        #the arrow configuration into one we've already seen.
        for perm in agroup:
            l = 0
            #this loop constucts the coloring with the correct arrows
            #for the permutation group to be applied to.
            for z in range(len(col)):
                if col[z][0] < 0:
                    temp_coloring_with_arroms[z] = deepcopy(col[z])
                elif col[z][0] >= 0:
                    temp_coloring_with_arroms[z] = deepcopy(col[z])
                    temp_coloring_with_arroms[z][0] = arrow_config[l]
                    l += 1

            new_arrow_config=[]
            arrow_perm = perm[1]
            color_perm = perm[0]
            new_coloring_with_arrows = []
            #here we use the permutations of the colors to permute the
            #colors for our configuration.
            for j in range(len(temp_coloring_with_arroms)):
                new_coloring_with_arrows.append(temp_coloring_with_arroms[color_perm[j]])
            #if there is an arrow on any lattice site it needs to
            #be updated. Otherwise just permute the colors.
            for site in new_coloring_with_arrows:
                if site[0] >= 0:
                    site[0] = arrow_perm[site[0]]
                    new_arrow_config.append(site[0])
            #if the new configuration has a smaller hash number than
            #the current configuration then we have seen it before and
            #it is not unique.
            permhash = _ahash(new_arrow_config,dim)
            if permhash < orighash:
                unique = False
                break
        #if none of the permutations have had a smaller hash number
        #than the current configuration then the configuration is
        #unique and we need to add it to the list.
        if unique == True:
            #this loop make the coloring with the arrows that will be
            #added to the list of unique configurations.
            coloring_with_arrows = [0]*len(col)
            i = 0
            for z in range(len(col)):
                if col[z][0] < 0:
                    coloring_with_arrows[z] = deepcopy(col[z])
                elif col[z][0] >= 0:
                    coloring_with_arrows[z] = deepcopy(col[z])
                    coloring_with_arrows[z][0] = arrow_config[i]
                    i += 1
                    
            arsurvivors.append(coloring_with_arrows)

    return(arsurvivors)

# arrow_concs is a method that returns the concentration string
# including the arrows
def arrow_concs(cList,aconcs):
    """Uses the concentrations of the atoms and the arrows to make a
    labeling for the system. It then sorts them to be in the correct
    order for the code.

    :arg cListr: an integer array the concentration of the colors 
    :arg aconcs: an array of floats of the fractional number of arrows
    for each color
    """

    aconcs = [int(cList[i]*aconcs[i]) for i in range(len(cList))]
    
    species = 1
    conc_w_arrows = []
    for i in range(len(aconcs)):
        na = aconcs[i]
        ns = cList[i]
        if ns >= na:
            while na > 0:
                conc_w_arrows.append([1,species])
                na -= 1
                ns -= 1
            while ns > 0:
                conc_w_arrows.append([-1,species])
                ns -= 1
        else:
            while ns > 0:
                conc_w_arrows.append([-1,species])
                ns -= 1
        species += 1

    conc_w_arrows = _col_sort(conc_w_arrows)

    return(conc_w_arrows)

def get_arrow_concs(params):
    """If the concentrations are being restricted then find the correct 
    arrow for each species included.

    :arg params: the lattice.in parameters.
    """
    a_concs = []
    if params["is_crestricted"]:
        for c in params["concs"]:
            if (len(c) == 4 and params["arrows"]):
                a_concs.append(c[3])
            else:
                a_concs.append(0)
    else:
        for i in range(params["nspecies"]):
            a_concs.append(0)
    return a_concs

def enum_sys(groupfile, concs, a_concs, num_wanted, HNF, params):
    """Enumerates a random subset of the unique structures that have the shape
    defined by the symmetry group and the specified concentration.

    :arg groupfile: path to the file containing the symmetry group.
    :arg concs: list of integer concentrations for each species.
    :arg a_concs: list of integer *arrow* concentrations for each species.
    :arg num_wanted: the number of structures to pick randomly from the enumerated
      list.
    :args HNF: the HNF for the system we're currently enumerating
    :args params: the dictionary of parameters read in from lattice.in
    """

    from phenum.grouptheory import get_full_HNF, get_sym_group
    from phenum.polyaburnside import polya
    from phenum.tree import brancher
    import phenum.io_utils as io

    decorations = arrow_concs(concs, a_concs)
    # get the symmetry group for this HNF. Assumes the group can be
    # found in the file labeled by (this_HNF)_sym_group.out
    if groupfile == None:
        sym_g = get_sym_group(params["lat_vecs"],params["basis_vecs"],get_full_HNF(HNF),3)
        agroup = []
        for i in range(len(sym_g.perm.site_perm)):
            agroup.append([sym_g.perm.site_perm[i],sym_g.perm.arrow_perm[i]])
    else:
        group = io.read_group(groupfile)
        # get symgroup from HNF and lat_vecs
        # add [0] to each element of the symmetry group
        agroup = [[g,[0]] for g in group]

    # we need to know the concentrations of the species with and
    # without arrows, we also need to know the number of arrows and
    # their species so we can undo the previous step later
    (n_arrows, arrow_types,concs_w_arrows) = how_many_arrows(decorations)

    # now find the number of unique arrangements using
    # polya
    if arrow_types != 0:
        # Since enumlib doesn't write the arrow group out we have to
        # recompute the group actions paired with their effects on the
        # arrows
        total = polya(concs_w_arrows, agroup, arrowings=arrow_types)
    else:
        total = polya(concs, agroup)

    # generate the random subset to be used
    if num_wanted < total:
        from random import shuffle
        subset = range(1, total+1) 
        shuffle(subset)
        subset = subset[0:num_wanted]
    else:
        from phenum.msg import warn
        warn("number of configurations requested exceeds the number of "
            "unique configurations available.")
        subset = []

    n_stabs = []
    # if we're doing a purely arrow enumeration then we don't need to
    # do the tree search but instead perform the final step of the
    # algorithm to find the possible unique displacements of the atoms
    if len(concs) == 1 and all(decorations) >=0:
        configs = []
        a_configs = add_arrows(decorations, agroup, 6)
        count = 1
        for config in a_configs:
            if count in subset:
                configs.append(config)
            count += 1
    else:
        if arrow_types != 0:
            configs = brancher(concs_w_arrows, agroup, decorations, 6, subset)
        else:
            configs = brancher(concs, agroup, decorations, 6, subset)

    # reduced_configs is the list of configurations with the
    # superperiodic configurations removed.
    reduced_configs = []
    # need to find a way to remove the superperiodic arrangements
    reduced_configs = configs                
    return reduced_configs

def _ahash(coloring,dim):
    """Produces a unique number for each possible configuration of
    arrows. This number is used to compare the order that the arrows
    occure in.

    :arg coloring: is a 2D integer array of the full coloring with colors and arrows	    
    :arg dim: the number of directions the arrows can point
    """
	
    narrow = 0
    for i in range(len(coloring)):
	narrow = narrow + coloring[i]*dim**i
    return(narrow)

#anum is a unique number that is associated with an array of arrows.
#num_of_arrows in the number of arrows that are in the array.
#dim is the number of directions the arrows can point.
def _ainvhash(anum,num_of_arrrows,dim):
    """Turns an arrow hash back into the array of arrow directions.

    :arg anum: the arrow hash number
    :arg num_of_arrows: the number of arrows in the system
    :arg dim: the number of directions the arrows can point
    """
    arrows = [0]*num_of_arrrows
    for i in range(num_of_arrrows):
	base = dim**(num_of_arrrows-1-i)
	arrows[num_of_arrrows-1-i] = anum//base
	anum = anum -base*(anum//base)
    return(arrows)



"""Methods for reading and writing enumeration and polya-counting results."""

"""Generates a random subset of possible structures weighted by the
Polya distribution for superstructures.
"""
def enum_data(cellsize, args, params):
    """Returns a list of dictionaries with the HNF, SNF, Left Transform
    and permutation group file names for the given cell size.
    """
    from os import path
    from numpy import loadtxt, array
    dirname = args["dataformat"].format(cellsize)
    result = []

    if path.isdir(dirname):
        matrices = loadtxt(path.join(dirname, "matrices"), int)
        limit = matrices.shape[0] if len(matrices.shape) > 1 else 1
        matrices = array([matrices]) if limit == 1 else matrices
        for i in range(limit):
            result.append({
                "SNF": matrices[i,1:4],
                "HNF": matrices[i,4:10],
                "L": matrices[i,10:],
                "group": path.join(dirname, "group.{}".format(matrices[i,0]))})
    else:
        from phenum.msg import warn, err
        from os import system
        from phenum.io_utils import write_struct_enum, which
        warn("No data for cell size {0:d} found at {1}. Creating new {2} files now using "
             "{3}.".format(cellsize, dirname,args["dataformat"],args["exec"]))
        if which(args["exec"]) != None:
            write_struct_enum(params)
            system(args["exec"])
            system('rm symops_enum_parent_lattice.out readcheck_enum.out fort.*')
            if path.isdir(dirname):
                matrices = loadtxt(path.join(dirname, "matrices"), int)
                limit = matrices.shape[0] if len(matrices.shape) > 1 else 1
                matrices = array([matrices]) if limit == 1 else matrices
                for i in range(limit):
                    result.append({
                        "SNF": matrices[i,1:4],
                        "HNF": matrices[i,4:10],
                        "L": matrices[i,10:],
                        "group": path.join(dirname, "group.{}".format(matrices[i,0]))})
            else:
                err("The executable {0} does not produce the needed input folders {1}, "
                    "or else your struct_enum.in doesn't agree with your {2} file. "
                    "\nPlease ensure that your struct_enum.in is correct. If it is then "
                    "please follow the instructions found to compile {0} in the "
                    "README.".format(args["exec"],args["dataformat"],args["lattice"]))
                exit()                            
        else:
            err("Could not find {} on your path. Please add it to your path "
                "if you have already\n compiled it. Otherwise please follow the "
                "instructions found in the README to download, make,\n and place the "
                "executable in your path.".format(args["exec"]))
            exit()
        
    return result   

def _read_struct_enum():
    """Reads in the struct_enum.out file and returns a dictionary with key
    for each cell size, and value a dictionary keyed by HNF with the number
    of unique structures.
    """
    i = 0
    structs = {}
    with open("struct_enum.out") as f:
        for line in f:
            i += 1                        
            if i > 17:
                vals = line.split()
                sN = int(vals[6])
                HNF = tuple(map(int, vals[11:17]))
                concs = tuple([vals[-1].count(c) for c in ["0", "1"]])
                if sN not in structs:
                    structs[sN] = {}
                if HNF not in structs[sN]:
                    structs[sN][HNF] = {}
                    structs[sN][HNF][concs] = 1
                else:
                    if concs in structs[sN][HNF]:
                        structs[sN][HNF][concs] += 1
                    else:
                        structs[sN][HNF][concs] = 1
    return structs

def _write_struct_summary(structs):
    """Writes the summary of unique structure counts by HNF and cell size
    to file for verification of polya.
    """
    from numpy import zeros
    for sN, HNFS in structs.items():
        out = open('enum_'+str(sN)+'.out', 'w+')
        out.write('{0: <28}'.format("# HNF"))
        first = sorted(HNFS.itervalues().next())
        for conc in first:
            out.write("{0: <10}".format(':'.join(map(str, conc))))
        out.write('{0: <10}\n'.format("Total"))

        conc_totals = {conc: 0 for conc in first}
        sHNFs = sorted(HNFS.values)
        for HNF in sHNFs:
            out.write("  {0: <26}".format(' '.join(map(str, HNF))))
            for conc in first:
                if conc in dHNF:
                    out.write("{0: <10d}".format(dHNF[conc]))
                    conc_totals[conc] += dHNF[conc]
                else:
                    out.write("{0: <10d}".format(0))
            out.write('{0: <10d}\n'.format(sum(dHNF.values())))
            
        out.write("# " + ''.join(['-' for i in range(len(first)*10 + 10 + 30)]) + '\n')
        out.write("{0: <28}".format(""))
        for conc in first:
            out.write("{0: <10d}".format(conc_totals[conc]))
        out.write("{0: <10d}\n".format(sum(conc_totals.values())))
        out.close()

def _distribute(cellsizes, ftype, n=None, dataformat="cells.{}"):
    """Returns a dictionary specifying how many of each cell shape, size and concentration
    to enumerate in order to obtain a grand total of 'n' unique cells, distributed according
    to the abundance of unique structures predicted by Polya. See the calling signature for
    method:distribution_fxn(). However, while that method does not guarantee the total 'n',
    this one does; if it is short a few at the end, it tops up the list according to the
    abundances across all cell sizes, shapes, etc.

    :arg cellsizes: a list of the cell sizes to include in the data set summary.
    :arg ftype: one of ['shape', 'conc', 'size', 'all']; determines the number of 
      arguments that the distr. function will accept and average over, thus controlling
      the granularity of the predictions.
    :arg n: the number of structure types to return, weighted by their polya distribution.
      If this value is None, then *all* structure types are returned with their unique
      number.
    :arg dataformat: the name of the directories to search for 'polya.out' files in.
    """
    (f, dataset, gtotal) = _distribution(ftype, None, None, cellsizes=cellsizes, dataformat=dataformat)
    if n > gtotal:
        from msg import warn
        warn("The number of unique structures you requested ({}) ".format(n) +
             "exceeds the total number of unique structures ({}). ".format(gtotal) +
             "The code will return *all* structures for the distribution.")
        n = gtotal

    #If they didn't specify an 'n', then we also return all the structures.
    if n is None:
        n = gtotal

    result = {}
    rtotal = [0]
    def vinsert(key, value, rtotal, n, result):
        if rtotal[0] < n: #Only update if we are deficient in structures.
            if value > 0:
                if key in result:
                    result[key] += value
                else:
                    result[key] = value
                rtotal[0] += value
            return True
        else:
            return False   
    
    def assign(relvals, f, rtotal, gtotal, n, result, ftype):
        """Recurses through the relative weights specified to fill the
        result dictionary until it has the exact number of structures
        requested.
        """
        if n == gtotal:
            for key, value, limit in relvals:
                vinsert(key, limit, rtotal, n, result)
        else:
            from numpy import round
            from math import modf
            from random import random
            from operator import itemgetter
            ids = {
                "all": lambda key: key,
                "shape": lambda key: key[0:2],
                "conc": lambda key: (key[0], key[2]),
                "size": lambda key: (key[0],)
            }
            #Sort the relative values according to their weights so that
            #the most abundant ones pick up the extras.
            relvals = sorted(relvals, key=itemgetter(1), reverse=True)
            recount = 0
            while rtotal[0] < n and recount < 4:
                delta = n-rtotal[0]
                for key, value, limit in relvals:
                    #If the abundance guarantees an integer part, use that, but then
                    #use a random number to decide who to keep for fractional parts.
                    #Eventually this should terminate unless all of the fractional
                    #pieces are super small; but even then, we should get what we want
                    #by the time we have iterated through the whole list.
                    rf, qf = modf(value if recount == 0 else f(delta, *(ids[ftype](key))))
                    retval = int(qf + (1 if random() < rf else 0))
                    if not vinsert(key, retval, rtotal, n, result):
                        break
                recount += 1

            if recount == 3 and rtotal[0] < n:
                from msg import warn
                warn("Reached maximum recursion limit in random assignment.")
                
    #First, compile a dictionary at the lowest level that has the weights
    #for the desired value of 'n'.
    relvals = []
    if ftype=="all":
        for size, data in dataset.items():
            for HNF, distr in data["distr"].items():
                for conc, limit in distr.items():
                    key = (size, HNF, conc)
                    value = f(n, size, HNF, conc)
                    relvals.append((key, value, limit))
    elif ftype == "shape":
        for size, data in dataset.items():
            for HNF, limit in data["stotals"].items():
                key = (size, HNF, None)
                value = f(n, size, HNF)
                relvals.append((key, value, limit))                
    elif ftype == "conc":
        for size, data in dataset.items():
            for conc, limit in data["ctotals"].items():
                key = (size, None, conc)
                value = f(n, size, conc)
                relvals.append((key, value, limit))
    elif ftype == "size":
        for size, data in dataset.items():
            limit = data["gtotal"]
            key = (size, None, None)
            value = f(n, size)
            relvals.append((key, value, limit))
    
    #This makes the selection according to the relative values and keeps
    #choosing, weighted by relative abundance, until we have the right
    #number of structures.
    assign(relvals, f, rtotal, gtotal, n, result, ftype)
    #Make sure we are returning exactly how many they asked for.
    if rtotal[0] > n:
        from msg import warn
        warn("More structures were returned than asked for, should not be possible.")

    from operator import itemgetter
    return result

def _distribution(ftype, dataset, gtotal, cast=float, cellsizes=None, dataformat="cells.{}"):
    """Returns the Polya-weighted distribution function for the specified cell sizes.
    If dataset is None, a dataset and total are loaded using the specified cell sizes.
    This dataset and the total number of structures it contains are returned as
    arguments 2 and 3 of the tuple. Depending on the value of 'ftype', it returns 
    a tuple (f, dataset, gtotal) where 'f' is a function function with arguments:

    'all':   f(n ,size, HNF, conc)
    'shape': f(n, size, HNF)
    'conc':  f(n, size, conc)
    'size':  f(n, size)

    where 'size' is the integer cell size; 'HNF' is the array of 6-elements defining
    the HNF matrix that is the cell shape; 'conc' is a list of concentrations such as
    [1,5] for a 6-atom cell; 'n' is the *grand* total number of structures desired from
    calling the function over all possible values of the various parameters passed to
    the distribution function.

    The value returned by the function is an integer specifying how many of that
    structure type should be selected according to the total distribution across
    all structures included in cellsizes.

    :arg dataset: the dictionary returned by method:_distribution_summary().
    :arg gtotal: the total number of unique structures corresponding to the
      dictionary 'dataset'.
    :arg ftype: one of ['shape', 'conc', 'size', 'all']; determines the number of 
      arguments that the distr. function will accept and average over, thus controlling
      the granularity of the predictions.
    :arg cellsizes: a list of the cell sizes to include in the data set summary.
    :arg dataformat: the name of the directories to search for 'polya.out' files in.
    """
    if dataset is None:
        if cellsizes is not None:
            (dataset, gtotal) = _distribution_summary(cellsizes, dataformat)
        else:
            from msg import err
            err("No dataset or cell sizes specified.")
            return None
        
    #We just need to sum up the total number of unique structures across all the
    #given cell sizes and then create a function that weights a total number of
    #desired structures by concentration and cell shape.
    ftotal = float(gtotal)
    if ftype == "all":
        f = lambda n, size, HNF, conc: cast(dataset[size]["distr"][tuple(HNF)][conc]/ftotal*n)
    elif ftype == "shape":
        f = lambda n, size, HNF: cast(dataset[size]["stotals"][tuple(HNF)]/ftotal*n)
    elif ftype == "conc":
        f = lambda n, size, conc: cast(dataset[size]["ctotals"][tuple(conc)]/ftotal*n)
    elif ftype == "size":
        f = lambda n, size: cast(dataset[size]["gtotal"]/ftotal*n)
    else:
        from msg import err
        err("The parameter {} is not a valid parameter for the distribution. Please use "
            " size, shape, conc, or all.".format(ftype))
        exit()
        
    return (f, dataset, gtotal)
        
def _distribution_summary(cellsizes, dataformat="cells.{}"):
    """Returns a dictionary that summarizes the Polya predictions for unique structures
    by cell size, shape and concentration for use by a distribution function. Returns
    a tuple (summary dict, grand total), where grand total is the total number of structures
    predicted for *all* cell sizes.

    :arg cellsizes: a list of the cell sizes to include in the data set summary.
    :arg dataformat: the name of the directories to search for 'polya.out' files in.
    """
    from os import path
    from numpy import loadtxt
    gtotal = 0
    dataset = {}
    for s in cellsizes:
        dirname = dataformat.format(s)
        source = path.join(dirname, "polya.out")
        if not path.isfile(source):
            from msg import err
            err("Cannot find polya distribution for size {} at {}".format(s, source))
            continue

        #First load the possible concentrations from the commented first line.
        with open(source) as f:
            headings = f.readline().split()
            concs = [tuple(map(int, h.split(":"))) for h in headings[2:-1]]
        polya = loadtxt(source, int)
        distr = {}
        for iHNF, HNF in enumerate(polya[0:-1,0:6]):
            distr[tuple(HNF)] = {tuple(c): v for c, v in zip(concs, polya[iHNF,6:-1])}
        stotals = {tuple(m): v for m, v in zip(polya[0:-1,0:6], polya[:-1,-1])}
        ctotals = {c: v for c, v in zip(concs, polya[-1, 6:-1])}
        dataset[s] = {
            "distr": distr,
            "concs": concs,
            "ctotals": ctotals,
            "stotals": stotals,
            "gtotal": polya[-1,-1]
        }
        gtotal += dataset[s]["gtotal"]
    return (dataset, gtotal)    

def _print_distribution(distr, filename=None, header=True, append=False):
    """Prints the specified distribution to screen or file.

    :arg distr: the distribution returned by method:distribute().
    """
    if filename is None:
        from msg import arb, cenum
        bysize = {}
        sfmt = " {0: <5d} | {1: <15} | {2: <5} | {3: <5d} "
        for key, value in distr.items():
            size, HNF, conc = key
            skey = sfmt.format(size, "" if HNF is None else ' '.join(map(str, HNF)),
                               "" if conc is None else ':'.join(map(str, conc)), value)
            cols = (cenum["cwarn"], cenum["cinfo"], cenum["cgens"], cenum["cokay"])
            if size in bysize:
                bysize[size].append((skey, cols))
            else:
                bysize[size] = [(skey, cols)]

        for size in sorted(bysize.keys()):
            for skey, cols in bysize[size]:
                arb(skey, cols, "|")
    else:
        #We don't worry about ordering it, just write them to file in whatever
        #order the keys are in.
        with open(filename, 'w' if not append else 'a') as f:
            if header:
                f.write("# {0: <28}  {1: <10}  {2}\n".format("HNF", "Conc.", "Number"))
            for key, value in distr.items():
                size, HNF, conc = key
                f.write("  {0: <28}  {1: <10}  {2:d}\n".format(' '.join(map(str, HNF)), ' '.join(map(str, conc)), value))

def make_enum_in(distribution,number=None,dataformat="cells.{}"):
    """Makes an enum.in file if the distrubiton type is all with the
    desired number of structures. Otherwise prints the distribution
    information to the screen for the user.

    :arg distribution: The parameters that the distribution is over
    ('shape', 'conc', 'size', 'all').
    :arg n: The number of structures or 'all'.
    :arg dataformat: The folder name for the cell sizes.

    """

    from os import listdir

    files = listdir(".")
    sizes = []
    form = dataformat.split(".")[0]

    ready = False
    for f in files:
        if form in f:
            sizes.append(int(f.split(".")[1]))
            ready = True

    if ready == False:
        from msg import err
        err("The files {} don't exist in this directory. You either need to run"
            " the -polya option or else navigate into the folder that contains "
            "the output {} folders.".format(dataformat))
        exit()

    distr = _distribute(sizes,distribution,n=number,dataformat=dataformat)
    if distribution.lower() == 'all':
        _print_distribution(distr,filename="enum.in")
    else:
        _print_distribution(distr)



"""The heart of the code is here. These methods build the tree and
check the canfigurations for uniqueness."""

def _coefficients(number_of_each_color,number_of_slots):
    """This method generates the radix numbers, or total number of
    combinations of each color, for a given aray.

    :arg number_of_each_color: is an integer array integer indicating the
    number of times each color is found in the array.
    :arg number_of_slots: an integer equal to the sum of the
    number_of_each_color array.
    """
    from numerics import binomial_coefficient

    #renames variables for convience and creates an empty array
    coe = []
    for i in range(0,len(number_of_each_color)):
    #Uses the binomial_coefficient method to build an array of radix numebrs
	coe.append(binomial_coefficient(number_of_slots,number_of_each_color[i]))
	number_of_slots -= number_of_each_color[i]
    return coe

def _hash(listi,color):
    """Generates a hash for each color in the ladeling returning a array
    of the hash for each color.
    
    :arg listi: An integer array that contains the labeling.
    :arg color: An integer array of the concentrations of the
    colors.

    For details on this method see: http://msg.byu.edu/papers/enum3.pdf
    """
    from numerics import binomial_coefficient

    m = len(listi)
    li = listi[::-1]
    rm1 = m - li.index(1)
    y = 0
    z = listi[:rm1].count(0)#m - listi[:rm1].count(1)
    li = listi
    for i in range(z):
	p0 = li.index(0) #next(x[0] for x in enumerate(li) if x[1] > j)
	k = li[p0:].count(1)
	li = li[p0+1:]
	y += binomial_coefficient(len(li),k-1)
    return y

def _invhash(branch,colors,n):
    """Turns a hash array into a labeling, i.e., undoes the hash()
    subroutine.
    
    :arg branch: an integer list of the radix number/hash array
    :arg colors: an integer array of the concentration of the colors
    :arg n: the length of the labeling array, i.e., the
    number of sites in the system

    For the details on this method see: http://msg.byu.edu/papers/enum3.pdf
    """

    from numerics import binomial_coefficient

    new = [0 for y in range(n)]
    coluse = 1
    for i in range(0,len(branch)):
        newi = [0 for y in range(n)]
        I = branch[i]
        t = colors[i]
        for l in range(n,0,-1):
            binom = binomial_coefficient(l-1, t-1)
            if binom <= I:
                I -= binom
            else:
                newi[n-l] = coluse
                t -= 1
        coluse += 1
        n -= colors[i]
        si = 0
        for s in range(0,len(new)):
            if new[s] == 0:
                new[s] = newi[si]
                si += 1
    return(new)

#color_list takes the configuration and returns a unique list of the
#colors used without duplicates.
def _color_list(col):
    """Finds the unique colors in a labeling.

    :arg col: an integer array of the labeling.
    """
    colors=[]

    tcol=col[:]
    for i in range(len(tcol)):
        if tcol[i] != 0:
            colors.append(tcol[i])
            for j in range(len(tcol)):
                if tcol[j] == tcol[i] and i != j:
                    tcol[j] =0
            tcol[i] =0
    return(colors)



def _perm(casei,colors,length,index,gen,stab,order,ast):
    """This method applies a cyclical permutation to an array to determine
    if it is unique. 
		
    :arg casei: the integer array hash for the current branch of
    the tree that will be permuted
    :arg colors: an integer array indicates the concentration of
    each color in the system
    :arg ast: the stabalizer for the final configuration that will
    be passed to the arrow permutation routine.
    :arg length: the total number of sites (colors) in the system
    :arg index: an integer that indicates which level of the tree
    we are in, i.e., which color is being added to the tree
    :arg gen: a 3D integer array that contains the permutation group
    :arg stab: a 3D integer array of the stabilizers for
    the previous level of this branch in the tree
    :arg order: is an integer array that keeps track of which
    configurations of the first level of the tree have already
    been seen

    The routin outputs the following:

    unique: an integer that indicates if the configuration is
    unique, a returned value of 0 is a unique configuration
    anything else is not
    st: a 2D integer array that stores the stabilizers for this
    configuration		   
    order and ast
    """
		
    #initialize unique to be 0 and arrays to be empty
    unique = 0
    if stab == 0:
	stab = []
    st = []
    ast = []
		
    # uses the invhash method from Inverse_radix_num to turn the id
    # number, casei, into an array
    li = list(_invhash(casei,colors,length))

    # if we're on the first level of the tree then we need to
    # follow a special procedure
    if index == 0:
	# first apply all the symmetry operations one at a time, a
	# is a single operation
	for a in gen:
	    # get the effect of the group on the colors, ingore the arrows
	    # i is the permutation of the colors for this operation
	    i = a[0]
	    lnew = []
	    # apply the permutation to get an equivalent configuration
	    for j in i:
		lnew.append(li[j])
	    lnewb = list(lnew)
                
	    # any color that was placed that isn't the color for the
	    # current level is set to zero
	    for y in range(0,len(lnew)):
		if lnew[y] != index + 1:
		    lnewb[y] = 0
		elif lnew[y] == index + 1:
		    lnewb[y] = 1

	    # hash from radix_num_generator turns the array back into
	    # an integer hash array
	    rtest = _hash(list(lnewb),colors)

	    # if the original hash is smaller than the new one we
	    # don't want to visit the new one again so we set its
	    # location in order to be -1
	    if casei[0] < rtest:
		order[rtest] = -1
            
	    # if the original hash is larger than the new one then
	    # this configuration isn't unique so we set unique to be 1
	    # and reset the other variables
	    elif casei[0] > rtest:
		unique += 1
		st = []
		ast = []
		break

	    # if the symmetry group didn't change this level
            # of the hash then it is a stabilizer and needs to
            # be saved
	    elif casei[0] == rtest:
		st.append(a)

	    # if the symmetry op changed nothing in the
	    # configuration then it is a stabilizer for the
	    # arrow level and needs to be saved
	    if li == lnew:
		ast.append(a)

    # if there is only a single stabilizer then the configuration is
    # unique and the stabilizer needs to be used for any later braches
    elif len(stab) == 1:
	st = stab
	ast = stab

    # if there are multiple stabilizers then we need to apply each of
    # them just like we needed to apply the entire group on the first
    # level of the tree
    elif len(stab) > 1:
	ast = []
        # first apply all the stabilizers one at a time, a is a single
        # operation
	for a in stab:
	    # get the effect of the group on the colors, ingore the
	    # arrows i is the permutation of the colors for this
	    # operation
	    i = a[0]
	    lnew = []
	    # apply the permutation to get an equivalent configuration
	    for j in i:
		lnew.append(li[j])
		lib= list(li)
	    lnewb = list(lnew)
	    lnewb2 = []

	    # any color that was placed that isn't the color for
	    # the current level is set to zero
	    for y in range(0,len(li)):
		if li[y] > index + 1:
		    lib[y] = 0
	    for y in range(0,len(lnew)):
		if lnew[y] > index + 1:
		    lnewb[y] = 0
	    for y in range(0,len(lnew)):
		if lnewb[y] == 0:
		    lnewb2.append(0)
		elif lnewb[y] == index + 1:
		    lnewb2.append(1)
                    
	    # if the the new and original configurations are the
	    # same then save the stabilizers
	    if lnewb == lib:
		st.append(a)
	    if li == lnew:
		ast.append(a)

	    # hash from radix_num_generator turns the array back
	    # into an integer hash array
	    rtest = _hash(list(lnewb2),colors)
	    # if the original hash is larger than the new one then
	    # this configuration isn't unique so we set unique to
	    # be 1 and reset the other variables
	    if casei[index] > rtest:
		unique = 1
		st = []
		ast = []
		break
            
	    # if the configuration isn't unique break from the
            # loop
	    if unique == 1:
		ast = []
		st = []
		break
    else:
	unique = 1
		
    return(unique,st,order,ast)


def brancher(concs,group,colors_w_arrows, dim, subset = []):
    """This routine navigates the tree and saves the unique configurations
    to an array survivors.
		
    :arg concs: the concentrations of the colors or atoms
    :arg group: the symmetry group for the system, including the
    effect on the arrows (displacement directions)
    :arg colors_w_arrows: an integer 2D array that indiciates
    which atoms are being displaced, i.e., where the arrows are.
    :arg dim: the number of directions the arrows can point
    :arg subset: an integer array of the subset of unique
    arrangements wanted
	
    The method returns the list of unique configurations and the
    number of stabilizers for the last level of the tree.

    survivors: is the 3D array of the unique arrangements of
    colors and arrows
    stob_len: the number of stabilizers for a given unique
    configuration
    """

    from phonons import how_many_arrows, add_arrows
    from copy import deepcopy

    # initial setup	

    # redifine the colors so that the arrows are treated like
    # their own color
    colors = _color_list(colors_w_arrows)
    # find the total number of atoms
    n = sum(concs)

    # Find the mixed radix number counter for the system
    C = _coefficients(concs,n)

    # count how many arrows there are
    (narrows,arrow_types,concs_w_arrows) = how_many_arrows(colors_w_arrows)
    # prepare the stabalizer array to be the appropriate size
    stabalizer = [0]*len(concs)
    
    # ast is the stabalizer for the arrow permutaiton order is used to
    # track which branchs of the first layer we have already seen so
    # we can skip over the ones we already know aren't unique
    ast = []
    order = {}
    for i in range(0,C[0]+1):
	order[i] = i

    # determine if we are finding a subset of doing a full enumeration
    if len(subset) > 0:
	use_subset = True
    else:
	use_subset = False
				
    # count is an integer counter used to keep track of where in
    # the total number of configurations we are.
    count = 1

    #create the all zero branch of lables
    #survivors is an array that stores the unique configurations
    from numpy import zeros
    branch = zeros(len(C))
    survivors = []

    #variables used for iteration and as test criteria for
    #navigating the tree
    # i keeps track of which level of the tree we are on
    # b0 lets us know when we've used all the needed colors/been
    # to all the relevant levels
    i = 0
    b0 = 0
		
    # Now we loop through the different possible hash arrays until
    # they have all been considered
    while branch[0] < C[0] and (len(survivors) < len(subset) or use_subset == False):
	# perm determines if the new array is unique, if yes unique =
	# 0, if no then unique = 1, perm also outputs the stabilizers
	# for each level, the order for the first level, and the
	# stabilizers for the arrow configurations
	(unique,stabalizer[i+1],order,ast) = _perm(branch,concs,n,i,group,stabalizer[i],order,ast)

	# if this array is unique then we may need to append it to the
	# list of survivors
	if unique == 0:
	    # if we've visited every level of the tree
	    # then we need to append the unique
	    # configurations
	    if b0 == 1 or i == 0:
		# if there are arrows we need to see
		# what their unique arrangements are
		# before saving the configurations
		if narrows > 0:
		    # first use invhash to turn the hash back to an array
		    brancht = list(_invhash(branch, concs, len(colors_w_arrows)))
		    # create a temporary copy of the branch.
		    tbrancht = []
		    for tt in range(len(brancht)):
			tbrancht.append(colors[brancht[tt]-1][1])
		    # make a coloring with arrows to be passed to the
		    # arrow permutiation code by adding arrows back
		    # into the array where needed.
		    for z in range(len(brancht)):
			brancht[z] = deepcopy(colors[brancht[z] -1])
		    # add_arrows from the phonon_brancher code returns
		    # the unique configurations with the unique arrow
		    # arrangements
		    arsurvivors = add_arrows(brancht,ast, dim)
		    #write the unique confgurations to file.
		    for z in arsurvivors:
			# if we aren't using a subset write everything
			# to file.
			if not use_subset:
			    survivors.append(z)
			# if we are then we only want to write the
			# random subset to file.
			else:
			    if count in subset:
				survivors.append(z)
			count += 1
		else:
		    # if there are no arrows just
		    # write the unique arrangement
		    # to file
		    if not use_subset:
                        tbranch = list(_invhash(branch, concs, len(colors_w_arrows)))
			survivors.append([[-1,leaf] for leaf in tbranch])
		    else:
			if count in subset:
                            tbranch = list(_invhash(branch, concs, len(colors_w_arrows)))
			    survivors.append([[-1,leaf] for leaf in tbranch])
		    count += 1

		    # if we aren't on the last contributing level
		    # of the tree then make sure b0 is 0
	    if i < len(branch) - 2:
		b0 = 0

	# all that follows dictates how we navigate the tree. This can
	# be a little tricky but does work at this time. Basically try
	# not to mess with the last few lines of this code if it can
	# be helped
								
	# if b0 is 0 then we need to move up to the next level of the
	# tree at this point
	if b0 == 0:
	    i += 1
	    # if we are now looking at the last level then set b0 to 1
	    # so that any unique configurations found will be saved
	    if i == len(branch) - 2:
		b0 = 1

	# root is a variable that tells us if we've filled this
	# level of the tree yet
	root = 0
	# if root = 0 and i>0 then we need to check to see if
	# we need to back track up the tree yet
	while root == 0 and i > 0:
	    # if we've used all the configurations for this level then
	    # we need to go back a level, if not then set root to 1 so
	    # that we keep moving forward through the tree
	    if branch[i] == C[i]-1:
		# first reset this level's number so that it will
		# increment properly the next time it's visited
		branch[i] = 0  
		# next reduce i by 1 to go back a level
		i -= 1
		# if i = 0 we're back to the bottom of the tree. We
		# need to reset the stabilizers, b0 and increment the
		# bottom layers index, branch[i], by 1
		if i == 0:
		    stabalizer = [0]*len(concs)
		    b0 = 0
		    branch[i] += 1
	    else:
		root = 1
                
	# if b0=1 then we need to increase this level index,
	# branch[i], by 1 to consider the next configuration
	if b0 == 1:
	    branch[i] += 1

	# test helps us check the first level for areas we can skip
	test = 0
	while test == 0:
	    # if we're on the first branch and the order
	    # for this configuration is negative then we
	    # skip it. Otherwise we need to consider it.
	    if i == 0 and order[branch[i]] == -1:
		branch[i] += 1
	    else:
		test = 1

    # done
    return survivors



"""Methods for taking an HNF and finding the site permutations and the
arrow permutations. All of these have been modified for python from
their original fortran implementations which can be found at:
https://github.com/msg-byu/enumlib/. Of these only get_rotation
get_rotation_perms_lists() has been modified in order to produce the
arrow group. The class ArrowPerm has also been added and implemented
inside the RotPermList class.
"""

import itertools
import numpy
from copy import deepcopy
import operator
import symmetry as sym 

class ArrowPerm(object):
    """ArrowPerm pairs a site permutation with an arrow permutation."""
    def __init__(self,site_perm = None,arrow_perm = None):
        """Initializes the ArrowPerm.
        :args site_perm: A 1D integer array containing a site permutation.
        :args arrow_perm: A 1D integer array containing an arrow permutation.
        """
        self.site_perm = site_perm
        self.arrow_perm = arrow_perm

class RotPermList(object):
    """RotPermList maintains the data structure for the rotations and
      permutations from the fortran code.
    """
    def __init__(self,nL = None,v =None,perm = None,RotIndx = None,arrows=None):
        """Initializes the RotPermList.
          :args nL: An integer indicating the number of operations
          :args v: A 2D integer array containing the lattice vectors
          :args perm: 
        """
        self.nL = nL
        self.v = v
        self.perm = ArrowPerm(site_perm = perm, arrow_perm=arrows)
        self.RotIndx = RotIndx

class opList(object):
    """opList maintains the data structure of the fixing operations form
      the fortran code.
    """
    def __init__(self,rot = None, shift = None):
        """Initializes the opList.
          :args v: A 3D array containing the rotations
          :args perm: An 2D array containing the shifts
        """
        self.rot = rot
        self.shift = shift

def _make_member_list(n):
    """Takes the length of three cyclic group and constructs the member
    list so that the permutations can be determined. Each member has
    three components, corresponding to the entries for each of the
    three cyclic groups.
    
    :args n: Integer array containing the diagonal elements of the SNF.
    """
    from operator import mul
    depth = reduce(mul,n,1)
    p = []
    for i in range(depth):
        p.append([0,0,0])
    for im in range(1,reduce(mul,n,1)):  # Loop over the members of the translation group
        p[im] = deepcopy(p[im-1]) # Start with the same digits as in the previous increment        
        p[im][2] = (p[im-1][2]+1)%n[2]  # Increment the first cyclic group
        if (p[im][2]==0):             # If it rolled over then
            p[im][1] = (p[im-1][1] +1) % n[1] # increment the next cyclic group
            if (p[im][1]==0):          # If this one rolled over too
                p[im][0] = (p[im-1][0]+1)%n[0] # Then increment the third one
    return p

def _find_permutation_of_group(g,gp):
    """
      :args g: Unpermuted groups
      :args gp: Permuted groups.
      :args perm: Permuted of gp.
    """
    n = len(g)
    perm = []
    skip = []
    for i in range(n):
        skip.append(False) # This is just for efficiency
    for im in range(n):
        for jm in range(n):
            if skip[jm]:
                continue # This is just for efficiency
            if gp[jm]==g[im]:
                perm.append(jm)
                skip[jm] = True
                break # don't keep looking if you already found the match
            
    return perm

def _is_equiv_lattice(lat1,lat2,eps):

    """This function determines whether two lattices are equivalent. That
      is, if one is an equal-volume derivative lattice of the
      other. It uses the idea of Santoro and Mighell
      (Acta. Cryst. 1972) that, for equivalent lattices, the
      transformation matrix that takes the vectors of str2 to str1 has
      determinant of 1 and all integer
      elements. [LV2]=[LV1][S]=&gt;S=LV1^-1*LV2. But I had to use abs
      on the determinant. This is not mentioned by Santoro and Mighell
      (issue of opposite handedness---which doesn't matter to me).

      :args lat1: 2D integer array containing the first lattice.
      :args lat2: 2D integer array containing the second lattice.
      :args eps: The finite precision tolerance.
    """
    is_equiv_lattice = False
    lat1inv = numpy.linalg.inv(numpy.array(lat1)).tolist()
    S = numpy.matmul(numpy.array(lat2),lat1inv).tolist()
    if (numpy.allclose(abs(numpy.linalg.det(numpy.array(S))),1.0,rtol=eps,atol=eps) and
        numpy.allclose(numpy.array(S),numpy.array([[int(round(S[i][j])) for j in range(3)] for i in range(3)]),rtol=eps,atol=eps)):
        is_equiv_lattice = True
    return is_equiv_lattice
        
def _get_sLV_fixing_operations(HNF,pLV,nD,rot,shift,dPerm,eps):
    """
      :args HNF:
      :args pLV:
      :args nD:
      :args rot:
      :args shift:
      :args dPerm:
      :args eps: Finite precision tolerance.
    """
    
    nRot = len(rot)

    degen_lattices = []
    for i in range(nRot):
        degen_lattices.append([[0,0,0],[0,0,0],[0,0,0]])
        
    cDegen = 0
    ic = 0 # Counter for the fixing operations
    tv = 0
    tmpOp_rot = []
    tmpOp_shift = []
    tv = []
    tIndex = [] # temp variables
    for iRot in range(nRot):  # Loop over each rotation
        thisRot = rot[iRot] # Store the rotation
        origLat = numpy.matmul(HNF,pLV).tolist()  # Compute the superlattice
        rotLat = numpy.matmul(numpy.array(origLat),numpy.array(thisRot)).tolist() # Compute the rotated superlattice
        if _is_equiv_lattice(rotLat,origLat,eps):
            # this operation fixes the lattice and should be recorded
            ic += 1
            tmpOp_rot.append(thisRot)
            tmpOp_shift.append(shift[iRot])
            tv.append(dPerm.v[iRot])
            tIndex.append(iRot)
            # Added by LN from here
        else:
            inList = False
            for iDegen in range(cDegen):
                if _is_equiv_lattice(degen_lattices[iDegen],rotLat,eps):
                    inList = True
                    break
          
            if not inList:
                degen_lattices[cDegen] = rotLat
                cDegen += 1
        # End of LN additions for degeneracy
        # Now we know which rotations fix the lattice and how many
        # there are so store them
    degeneracy = cDegen
    # Allocate the storage for them
    fixOp = opList(tmpOp_rot,tmpOp_shift) # Stuff the rotations into the permanent array
    if nD > 1:
        rotPerm = RotPermList(v=tv,RotIndx=tIndex)
    else:
        rotPerm = RotPermList(v=[tv],RotIndx=tIndex)

    return(fixOp,rotPerm,degeneracy)

def _map_dvector_permutation(rd,d,eps,n):
    """
      :args rd: 2D array of rotated basis vectors
      :args d: 2D array of original basis vectors
      :args eps: Finite precision tolerance.
      :args n: number of basis vectors
    """

    found = []
    for i in range(len(rd)):
        found.append(False)
    nD = len(rd) # of d-vectors
    RP = []
    for iD in range(nD):
        for jD in range(nD):
            if found[jD]:
                continue
            if numpy.allclose(numpy.array(rd[iD]),numpy.array(d[jD]),atol=eps,rtol=eps):
                RP.append(jD)
                found[jD] = True
                break

    if len(RP) != len(d):
        print("d-vector didn't permute in map_dvector_permutation "
              "This usually means that the d-set from the input structure and the d-set"
              " from the struct_enum.out have a different origin or don't live in the same"
              " unit cell. This probably isn't your fault---the code should overcome this."
              ,RP)
        exit()
    return(RP)

        
def _find_minmax_indices(invec):
    """Finds the indices corresponding to the minimum and maximum values
      in an integer vector.

      :args invec: The input integer array.
    """

    vec = [abs(i) for i in invec]
    
    this_min = invec.index(min([i for i in vec if i > 0]))

    rvec = [i for i in reversed(vec)]
    
    this_max = 2 - rvec.index(max(rvec))

    return (this_min,this_max)

def SmithNormalForm(HNF):
    """Finds the Smith Normal Form (SNF) of an HNF as well as the left and
      right transforms.
      
      :args HNF: The integer matrix HNF for which the SNFs are to be
            found
    """

    if numpy.linalg.det(numpy.array(HNF)) < 1:
        print("SmithNormalForm routine failed because the input matrix had a negative determinant")
        exit()
    
    A = [[0,0,0],[0,0,0],[0,0,0]]
    M = deepcopy(HNF)
    B = [[0,0,0],[0,0,0],[0,0,0]]

    for i in range(3):
        A[i][i] = 1
        B[i][i] = 1

    j = 0
    itCnt = 0

    stop_loop = False
    while not stop_loop:
        itCnt += 1
        if (itCnt >=100):
            print("Bad progromming in SmithNormalForm")
            exit()

        while (3-[M[0][j],M[1][j],M[2][j]].count(0)) > 1:
            (minidx,maxidx) = _find_minmax_indices([M[0][j],M[1][j],M[2][j]])
            minm = M[minidx][j]
            mult = M[maxidx][j]/minm

            M[maxidx] = map(operator.sub,M[maxidx],[mult*i for i in M[minidx]])
            A[maxidx] = map(operator.sub,A[maxidx],[mult*i for i in A[minidx]])

            if numpy.dot(numpy.dot(numpy.array(A),numpy.array(HNF)),numpy.array(B)).any() != numpy.array(M).any():
                print("ROW: Transformation matrices didn't work")
                exit()

        if M[j][j] == 0:
            maxidx = [abs(M[0][j]),abs(M[1][j]),abs(M[2][j])].index(max([abs(M[0][j]),abs(M[1][j]),abs(M[2][j])]))
            tmprow = deepcopy(A[j])
            A[j] = deepcopy(A[maxidx])
            A[maxidx] = tmprow

            tmprow = deepcopy(M[j])
            M[j] = deepcopy(M[maxidx])
            M[maxidx] = tmprow    
        if numpy.dot(numpy.dot(numpy.array(A),numpy.array(HNF)),numpy.array(B)).any() != numpy.array(M).any():
            print("ROWSWAP: Transformation matrices didn't work")
            exit()

        while (3-M[j].count(0)) >1:
            (minidx,maxidx) = _find_minmax_indices(M[j])
            minm = M[j][minidx]
            mult = M[j][maxidx]/M[j][minidx]

            for i in range(3):
                M[i][maxidx] = M[i][maxidx]-mult * M[i][minidx]
                B[i][maxidx] = B[i][maxidx]-mult * B[i][minidx]

            if numpy.dot(numpy.dot(numpy.array(A),numpy.array(HNF)),numpy.array(B)).any() != numpy.array(M).any():
                print("COLS: Transformation matrices didn't work")
                exit()

            if M[j][j] < 0:
                for i in range(3):
                    M[i][j] = -M[i][j]
                    B[i][j] = -B[i][j]
            else:
                maxidx = [abs(i) for i in M[j]].index(max([abs(i) for i in M[j]]))
                for i in range(3):
                    tmp = B[i][j]
                    B[i][j] = B[i][maxidx]
                    B[i][maxidx] = tmp 
                    tmp = M[i][j]
                    M[i][j] = M[i][maxidx]
                    M[i][maxidx] = tmp

        if ((3-M[j].count(0)) >1) or ((3-[M[0][j],M[1][j],M[2][j]].count(0)) >1):
            continue
        if numpy.dot(numpy.dot(numpy.array(A),numpy.array(HNF)),numpy.array(B)).any() != numpy.array(M).any():
            print("COLSWAP: Transformation matrices didn't work")
            exit()

        Ldiv = [[True,True],[True,True]]
        for i in range(1,3):
            for k in range(1,3):
                Ldiv[i-1][k-1] = (M[i][k]%M[0][0] == 0)

        if j == 0 and any(Ldiv) == False:
            local = [[0,0],[0,0]]
            for i in range(1,3):
                for k in range(1,3):
                    local[i][k] = abs(M[i][k]%M[0][0])
            nondividx = local.index(max(local))
            M[0] = map(operator.add,M[0],M[nondividx+1])
            A[0] = map(operator.add,A[0],A[nondividx+1])
            continue

        if j == 1:
            if M[2][2]%M[1][1] != 0:
                M[1] = map(operator.add,M[1],M[2])
                A[1] = map(operator.add,A[1],A[2])
                continue
        else:
            j = 1
            continue
        if j == 1 and (M[2][1] != 0 or M[1][2] != 0):
            continue
        stop_loop = True
         
    if M[2][2] < 0:
        for i in range(3):
            M[i][2]=-M[i][2]
            B[i][2]=-B[i][2]
    
    if numpy.dot(numpy.dot(numpy.array(A),numpy.array(HNF)),numpy.array(B)).any() != numpy.array(M).any():
        print("END: Transformation matrices didn't work.")
        exit()
    check = [M[0][1],M[0][2],M[1][0],M[1][2],M[2][0],M[2][1]]
    if any(check) != 0:
        print("Not diagonal")
        exit()
    if M[1][1] % M[0][0] != 0 or M[2][2] % M[1][1] != 0:
        print("SNF conditions not met")
        exit()

    return(M,A,B)

def _get_dvector_permutations(par_lat,bas_vecs,LatDim,eps):
    """This routine applies the symmetry operations of the parent lattice
      to the interior points (i.e., the d-set) to see which ones are
      equivalent. Labelings of the lattice points that are contain
      permutations only of labels on equivalant sites are physically
      equivalent and therefore redundant. We use these permutations to
     eliminate those duplicate labelings

      :args pLV: A 2D integer array containing the parent lattice vectors
      :args bas_vacs: A 2D integer array containing the basis vectors for the cell
      :args LatDim: 2 if a 2D case 3 if 3D
      :args eps: finite precesion tolerance
    """

    nD = len(bas_vecs)
    aTyp = []
    for i in range(nD):
        aTyp.append(1)

    (rot,shift)= sym.get_spaceGroup(par_lat,aTyp,bas_vecs)

    if LatDim==2:
        (rot,shift) = _rm_3D_operations(par_lat,rot,shift,eps)

    nOp = len(rot)
    
    inv_par_lat = numpy.linalg.inv(numpy.array(par_lat)).tolist()
    
    nL_temp= nOp  # Number of operations that fix the parent

    # lattice (but may permute the d-vectors)
    v_temp = []
    perm_temp = []
    for iOp in range(nOp): # Try each operation in turn and see how the
        # d-vectors are permuted for each
        # Rotate each d and add the shift
        if len(bas_vecs) == 1:
            temp_b = bas_vecs[0]
        else:
            temp_b = bas_vecs
        rd_rot = numpy.matmul(temp_b,rot[iOp]).tolist()
        if nD > 1:
            rd_shift = [shift[iOp]]*nD
            rd =  [[rd_rot[i][j] + rd_shift[i][j] for j in range(len(rd_rot[i]))] for i in range(len(rd_rot))]
            
        else:
            rd = [rd_rot[i] + shift[iOp][i] for i in range(len(rd_rot))]
        tRD = deepcopy(rd)
        if nD > 1:
            for iD in range(nD):
                rd[iD] = sym.bring_into_cell(rd[iD],inv_par_lat,par_lat,eps)
        else:
            rd = sym.bring_into_cell(rd,inv_par_lat,par_lat,eps)

        # The v vector is the vector that must be added (it's a lattice
        # vector) to move a rotated d-vector back into the parent cell.
        if nD > 1:
            v_temp.append([[rd[i][j] - tRD[i][j] for j in range(len(rd[i]))] for i in range(len(rd))])
        else:
            v_temp.append([rd[i] - tRD[i] for i in range(len(rd))])
        if nD > 1:
            perm_temp.append(_map_dvector_permutation(rd,bas_vecs,eps,nD))
        else:
            perm_temp.append(_map_dvector_permutation([rd],[bas_vecs],eps,nD))

    dRPList = RotPermList(nL=nL_temp,v=v_temp,perm=perm_temp)

    # I don't think we should reduce this list to a unique one. Some
    # rotations that don't permute the d's could still permute the
    # g's. So we have to keep all the d permutations, even if they
    # look redundant here.

    return(dRPList)
    
def _get_rotation_perms_lists(A,HNF,L,SNF,Op,RPlist,dperms,eps):
    """For each HNF, we have a list of the operations (rotations + shifts,
      if present) that leave the superlattice fixed. Given this set of
      fixing operations, make a list of the permutations on the
      d-vectors (interior points of the multilattice) effected by the
      rotations. Then sort the HNFs into blocks that have the same
      rotation permutations.

      :args A: Lattice vectors of the primary lattice (parent lattice).
      :args HNF: HNF matrix.
      :args L: Left transforms matrix.
      :args SNF: SNF matrix.
      :args Op: A list of symmetry ops (rots and shifts) for the parent multilattice.
      :args RlList:A list of permutations effected by the Ops.
      :args dperms:
      :args eps: Finite precision tolerance.d
    """

    # Index of the superlattices; Number of d-vectors in d set
    n = int(round(numpy.linalg.det(numpy.array(HNF))))

    nD = numpy.array(dperms.perm.site_perm).shape[-1]
    # nD = len(RPlist.v[0])

    skip = []
    gp = []

    arrowg = [[0,0,-1],[0,-1,0],[-1,0,0],[1,0,0],[0,1,0],[0,0,1]]
    
    tg = []
    perm = []
    ident = []
    identT = []
    tperms_perm = []
    temp_rperms_perm = []
    identT = [list(i) for i in zip(*[iter(range(0,n*nD))]*n)]  # we could combine
    # those two lines, but gfortran
    ident  = numpy.transpose(identT).tolist()            # does some strange things then.
    RPlist.nL=0  # initialize the number

    # Make the group member list for the first SNF
    diag = [SNF[0][0],SNF[1][1],SNF[2][2]]
    g = _make_member_list(diag)
    # arrowg = make_arrow_list(diag)
    Ainv = numpy.linalg.inv(A)
    L = numpy.transpose(L).tolist()
    # Make the transform matrices for taking the g's and rotating them
    Tinv = numpy.matmul(L,Ainv).tolist()
    T = numpy.linalg.inv(Tinv).tolist()
       
    nOp = len(Op.rot)

    naOp = len(arrowg)
    
    temp_rperm_perm = []
    temp_arrow_perm = []

    for iOp in range(nOp): # For each rotation, find the permutation
        dap = []
        dgp = []
        for i in range(n):
            tt = []
            for j in range(nD):
                tt.append(0)
            dgp.append(tt)

        for i in range(naOp):
            dap.append(0)

        for iD in range(nD): # Loop over each row in the (d,g) table
            if nD == 1:
                temp1 = [RPlist.v[iD][iOp]]*n
            else:
                temp1 = [RPlist.v[iOp][iD]]*n

            temp2 = numpy.transpose(numpy.matmul(numpy.matmul(Op.rot[iOp],T),numpy.transpose(g))).tolist()
            # temp2a = numpy.transpose(numpy.matmul(numpy.matmul(Op.rot[iOp],T),numpy.transpose(arrowg))).tolist()
            rag = numpy.transpose(numpy.matmul(Op.rot[iOp],numpy.transpose(arrowg))).tolist()
            rgp = numpy.transpose(numpy.matmul(Tinv,numpy.transpose([[temp1[i][j]+temp2[i][j] for j in range(len(temp1[i]))] for i in range(len(temp1))]))).tolist()
            # rag = numpy.transpose(numpy.matmul(Tinv,numpy.transpose([[temp1a[i][j]+temp2a[i][j] for j in range(len(temp1a[i]))] for i in range(len(temp1a))]))).tolist()
            temp_gp = [[int(round(rgp[i][j])) for j in range(len(rgp[i]))] for i in range(len(rgp))] # Move the rotated group into an integer array
            temp_ag = [[int(round(rag[i][j])) for j in range(len(rag[i]))] for i in range(len(rag))]
            if not numpy.allclose(rgp,temp_gp,rtol=eps,atol=eps):
                print("Transform left big fractional parts")
                exit()

            gp = temp_gp
            ag = temp_ag
            temp_diag = [diag]*n
            gp = [[gp[i][j] % temp_diag[i][j] for j in range(len(gp[i]))] for i in range(len(gp))] # Mod by each entry of
            # the SNF to bring into group Now that the rotated group
            # is known, find the mapping of the elements between the
            # original group and the permuted group. This is the
            # permutation.
            skip = []
            for mm in range(n):
                skip.append(False) # This is just for efficiency
            for im in range(n):
                for jm in range(n):
                    if skip[jm]:
                        continue # Skip elements whose mapping is already known
                    if gp[jm] == g[im]: # these elements
                        # map to each other The list of operations that fix
                        # the superlattice are a subset of those that fix
                        # the parent lattice. RotIndx stores the indicies
                        # of the parent lattice operations in a list with
                        # as many entries as supercell fixing operations.

                        # dperms%perm stores a list of d-vector
                        # permutations, one permutation (an nD list) for
                        # each operation in the parent lattice symmetries
                        OpIndxInSuperCellList = RPlist.RotIndx[iOp]
                        RowInDxGTable = numpy.transpose(dperms.perm.site_perm).tolist()[iD][OpIndxInSuperCellList]
                        dgp[im][RowInDxGTable] = jm+iD*n
                        skip[jm] = True
                        break
            # do the some thing for the arrows
            skip = []
            for mm in range(naOp):
                skip.append(False)
            for im in range(naOp):
                for jm in range(naOp):
                    if skip[jm]:
                        continue
                    if ag[jm] == arrowg[im]:
                        dap[im] = jm
                        skip[jm] = True
                        break

            if dgp.count(0) > 1 or dap.count(0) > 1:
                print("(d,g)-->(d',g') mapping failed in get_rotation_perm_lists")
                exit()

        # Now we have the (d',g') table for this rotation. Now
        # record the permutation
        temp_rperms_perm.append(numpy.transpose(dgp).reshape(nD*n).tolist()) # store
        temp_arrow_perm.append(dap)
        # permutation in the "long form"
        temp_rperms_nL = nOp
    # nomenclature:
    # N+t = rotation (N) + fractional translation (t)  (me bethinks....)
    # r = lattice translation

    # Now that we have the permutations that are effected by N+t
    # type of rotations (for each N in the point group), we need to
    # compose them with the permutations effected by lattice
    # translations, (parent) lattice translations contained inside the
    # supercell (N+t+r, in Rod's nomenclature). Only when we have these
    # two lists, and compose them, do we have a guarantee that the
    # resulting list is actually a group. For efficiency, we reduce the
    # N+t list (remove duplicates). Rod claims that the compositions
    # (with the translations) will not have duplicates.
    if len(temp_rperms_perm) > 1 and len(temp_arrow_perm) == 0:
        temp_rperms_perm.sort()
        temp_rperms_perm = list(temp_rperms_perm for temp_rperms_perm, _ in itertools.groupby(temp_rperms_perm))
    # The rotations permutations list is now in "alphabetical"
    # order and contains no duplicates
    
    # To get the permutations effected by the r's, we don't need
    # the r's. We can merely take the member list, the group, and
    # add each element of the group to the group itself and see
    # what permutation happens. (This part is somewhat redundant
    # with make_translation_group in labeling_related module.)
    for ig in range(n): # The number of r's inside the superlattice (the
        # translation perms) is the same as the index n
        temp_g = [g[ig]]*n
        tg = [[g[i][j]+temp_g[i][j] for j in range(len(g[i]))] for i in range(len(g))] # Add the element to the group
        temp_diag = [diag]*n
        tg = [[tg[i][j]%temp_diag[i][j] for j in range(len(tg[i]))] for i in range(len(tg))] # mod by the SNF entries to
        # bring it back to the "primitive" representation
        perm = _find_permutation_of_group(g,tg)
        temp_ident = []
        trans_ident = numpy.transpose(ident).tolist()
        for il in range(len(numpy.transpose(ident).tolist())):
            temp_ident.append([trans_ident[il][i] for i in perm])
        tperms_perm.append(numpy.reshape(temp_ident,(n*nD)).tolist())

    RPlist_perm_sites = []
    RPlist_perm_arrows = []
    RPlist_nL = len(temp_rperms_perm)*n
    for it in range(n): # Loop over translation perms (r type)
        for iOp in range(len(temp_rperms_perm)): # Loop over unique rotation
            # perms (N+t type) Form the permutation effected by
            # composing the iOp-th one with the it-th one
            RPlist_temp = [tperms_perm[it][i] for i in temp_rperms_perm[iOp]]
            RPlist_perm_sites.append(RPlist_temp)
            RPlist_perm_arrows.append(temp_arrow_perm[iOp])
            # ^--- Having gotten both the rotations and the
            # translation in the sections above (sort_permutations_list
            # etc...), the "operators" of the rotation (one of them is
            # R_i) and the translations (one of them is T_j) are now
            # "scrambled", i.e., (T_j) o (R_i).  Since the *first*
            # Rotation R_1 = Id, the *first* entries in the
            # RPlist(iH)%perm are equivalent to pure translations only.
            # The following entries are a combination of both.

    RPlist.perm.site_perm = RPlist_perm_sites
    RPlist.perm.arrow_perm = RPlist_perm_arrows
    RPlist.n = RPlist_nL
    return (RPlist)
    
    
def get_sym_group(par_lat,bas_vecs,HNF,LatDim):
    """Generates the symmetry group for a given lattice and HNF

    :args par_lat: a 2D integer array that contains the parrent
    lattice vectors
    :args bas_vecs: The atomic basis for the lattice
    :args HNF: The HNF that defines the supercell
    :args LatDim: 2 if a 2D case 3 if 3D
    """

    eps = 1E-10
    ParRPList = _get_dvector_permutations(par_lat,bas_vecs,LatDim,eps)

    aTyp = []
    for i in range(len(bas_vecs)):
        aTyp.append(1)
        
    (sgrots,sgshifts) = sym.get_spaceGroup(par_lat,aTyp,bas_vecs,eps)
    (fixing_ops,RPList,degeneracy) = _get_sLV_fixing_operations(HNF,par_lat,len(bas_vecs),sgrots,
                                                    sgshifts,ParRPList,eps)
    
    (SNF,L,R) = SmithNormalForm(HNF)

    symm = _get_rotation_perms_lists(par_lat,HNF,L,SNF,fixing_ops,RPList,ParRPList,eps)

    return(symm)

#a_group take the set of translations of the lattice and the set of
#rotations of the lattice paired with their effect on the arrows of
#the lattice. It makes a direct product of the transformations and the
#rotations and pairs them with the combination of th arraws. If the
#resultant operation is not in the group it then adds it to the group.
def a_group(trans,rots):
    """This subroutine combines that translations of the lattice with the
    rotatians of the lattice.

    :arg trans: a 2D integer array where each row is an translation
    of the lattice
    :arg rots: a 3D integer array. Each row's first entry is the
    site permutations and the second entry is the arrow
    permutations.
    """
    groupi = []
    for i in trans:
        for j in rots:
            c = []            
            c.append([j[0][i[l]] for l in range(0,len(i))])
            c.append(j[1])
            if c not in groupi:
                groupi.append(c)

    return(groupi)

def a_group_gen(trans,rots):
    """This subroutine combines that translations of the lattice with the
    rotatians of the lattice for the generators of the system.

    :arg trans: a 2D integer array where each row is an translation
    of the lattice
    :arg rots: a 3D integer array. Each row's first entry is the
    site permutations and the second entry is the arrow
    permutations.
    """
    groupi = []
    for i in trans:
        for j in rots:
            c = []            
            c.append([j[0][i[l]] for l in range(0,len(i))])
            c.append(j[1])
            if c not in groupi:
                groupi.append(c)

    groupi = _group(groupi)
    
    return(groupi)

def _group(gen):
    """This subroutine takes the generators of the group then uses them
    to form the entire group.

    :arg gen: a 2D integer array of the generators of the group
    where each row is a seperate generator
    """
    # q is a counter to track how many group elements we've found
    q = 1
    # the final output group
    groupi = []
    # loop over the generators
    for i in gen:
        # a second set of generators used to ensure that a group
        # action doesn't act an its self
        gen2 = []      
        # if k is not the ith element of the generators add it to the
        # set gen2
        for k in gen:
            if k != i:
                gen2.append(k)
        # loop over all except the ith generator
        for j in gen2:
            # handle the site group and arrow group independently
            trans = [j[0][i[0][l]] for l in range(0,len(i[0]))]
            rots = [j[1][i[1][l]] for l in range(0,len(i[1]))]
            # c is the combination of the trans and rots, if it's not
            # in groupi then it should be added.
            c = [trans,rots]
            if c not in groupi:
                groupi.append(c)

            # for each new group element, c, we should also see if
            # the ith group element acting on it will make another
            # new group element.
            trans = [c[0][i[0][l]] for l in range(0,len(i[0]))]
            rots = [c[1][i[1][l]] for l in range(0,len(i[1]))]
            # d is the group action after the ith element has acted on
            # c, if it's not in groupi then add it.
            d = [trans, rots]
            if d not in groupi:
                groupi.append(d)
            # until d and c are the same again we want to apply the
            # ith group element to d so that we've found all possible
            # group elements for the combination
            while d != c:
                trans = [d[0][i[0][l]] for l in range(0,len(i[0]))]
                rots = [d[1][i[1][l]] for l in range(0,len(i[1]))]
                d = [trans,rots]
                # increment q because we found a new group element
                q += 1
                # every time d is not in group i add it
                if d not in groupi:
                    groupi.append(d)
    # new is a tracker that is 0 while we are still finding new group
    # elements and 1 when no new ones are found
    new = 0

    #Here each generator is applied to the group over and over until no new group
    #elements are found
    while new == 0:
        # a second set to store any additional group elements
        group2 = []
        for i in gen:
            for h in groupi:
                # apply the ith generator to the hth element of groupi
                # from above to search for new group elements
                trans = [h[0][i[0][l]] for l in range(0,len(i[0]))]
                rots = [h[1][i[1][l]] for l in range(0,len(i[1]))]
                
                d = [trans,rots]
                # if do isn't in groupi or in group2 then add it to
                # group2
                if d not in groupi:
                    if d not in group2:
                        group2.append(d)
        # add anything that is in group2 to groupi
        if len(group2) > 0:
            groupi.extend(group2)
        else:
        # if no new group elements were found then we'v found the
        # whole group and new is set to 1
            new = 1
        
    return(groupi)

def get_full_HNF(HNF):
    """Turns the reduced HNF to a full HNF matrix.
    
    :args HNF: The 1D integer numpy array that contains all the lower
    diagonal entries of the HNF matrix.
    """

    temp_HNF = HNF.tolist()
    full_HNF = [[temp_HNF[0],0,0],[temp_HNF[1],temp_HNF[2],0],temp_HNF[3:]]

    return full_HNF


def _rm_3D_operations(aVecs,sgrots,sgshifts,eps):
    """This subroutine removes operations that are 3 dimmensional.

    :args aVecs: 2D integer array of the primitive real space lattice vectors
    :args sgrots: 3D integer array containing the space group rotations.
    :args sgshifts: 2D integer array containing the space group shifts.
    :args eps: Finite precisions tolerance.
    """
  
    if not numpy.allclose(numpy.array(aVecs[0][1:3]),0.0,rtol=eps,
                          atol=eps) or not numpy.allclose(numpy.array(aVecs[2][0:2])
                                                          ,0.0,rtol=eps,atol=eps):
        print("Error in rm_3d_operations: only allowed for primitive vectors x00,0xx,0xx")
        exit()
    
    nRot = len(sgrots)
    irot = 0
    tSGrots = []
    tSGshifts = []
    for i in range(nRot):
        if (numpy.allclose(numpy.array(sgrots[i][0][1:3]),0.0,rtol=eps,atol=eps) and
            numpy.allclose(numpy.array([sgrots[i][1][0],sgrots[i][2][0]]),0.0,rtol=eps,atol=eps)
            and numpy.allclose(abs(sgrots[i][0][0]),1.0,rtol=eps,atol=eps)):
            # this operation is "2D"         
            irot += 1
            tSGrots.append(sgrots[i])
            tSGshifts.append(sgshifts[i])

    nRot = irot

    return(tSGrots,tSGshifts)



#!/usr/bin/env python
try:
    from setuptools import setup
    args = {}
except ImportError:
    from distutils.core import setup
    print("""\
*** WARNING: setuptools is not found.  Using distutils...
""")

from setuptools import setup
try:
    from pypandoc import convert
    read_md = lambda f: convert(f, 'rst')
except ImportError:
    print("warning: pypandoc module not found, could not convert Markdown to RST")
    read_md = lambda f: open(f, 'r').read()

setup(name='phenum',
      version='1.4.1',
      description='Enumeration of symmetrically unique derivative superstructures of crystals.',
      long_description=read_md('README.md'),
      author='Wiley S Morgan',
      author_email='wiley.s.morgan@gmail.com',
      url='https://github.com/wsmorgan/phonon-enumeration',
      license='MIT',
      install_requires=[
          "numpy",
          "termcolor",
          "pyparsing",
      ],
      packages=['phenum'],
      scripts=['phenum/enumeration.py'],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Science/Research',
          'Natural Language :: English',
          'License :: OSI Approved :: MIT License',          
          'Operating System :: MacOS',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Topic :: Scientific/Engineering :: Physics',
      ],
     )



"""Methods used for generating the symmetry group. All of these
methods were adapted to python from their original fortran
implementations which can be found at:
https://github.com/msg-byu/symlib
"""

from copy import deepcopy
import numpy
import math

def get_concs_for_size(size,nspecies,res_concs,nB,concs):
    """Gets the concentration ranges for the atoms within the cells of
    certain sizes given the constraints provided such as
    concentration restrictions and the presence of arrows. Code
    rewritten from the get_conetration_list subroutine of:
    https://github.com/msg-byu/enumlib/blob/master/src/derivative_structure_generator.f90  

    :arg size: the cell size in integer form
    :arg nspecies: the number of atomic species in the system
    :arg res_concs: a logical that indicates of the concentrations
    are being restricted
    :arg nB: the number of basis vectors being used
    :arg concs: an 2D integer array that contains the concentration
    ranges for each atom in the system
    """

    eps = 1E-10
    
    from itertools import product
    if res_concs == True:
        denom = concs[0][2]
        volTable = []
        for atom in concs:
            minc = min(atom[0:2])
            maxc = max(atom[0:2])
            volTable.append([int(math.floor(float(minc)/denom*size*nB)),int(math.ceil(float(maxc)/denom*size*nB)),size*nB])

        n = volTable[0][2]
        digit = [volTable[i][1]-volTable[i][0] for i in range(len(volTable))]
        digCnt = [0*i for i in range(len(volTable))]
        k = len(volTable)

        label = []
        minv = []
        maxv = []
        for i in range(k):
            label.append(range(volTable[i][0],volTable[i][1]+1))
            minv.append(float(min([concs[i][0],concs[i][1]]))/concs[i][2])
            maxv.append(float(max([concs[i][0],concs[i][1]]))/concs[i][2])

        a = [label[i][0] for i in range(len(label))]
        
        cc = 0
        cList = []
        done = False
        while done == False:
            if sum(a) == n:
                conc = []
                for i in range(len(a)):
                    conc.append(a[i]/float(n))
                if not ((any(conc[i] < (minv[i]-eps) for i in range(len(minv)))) or (any(conc[i] > (maxv[i]+eps) for i in range(len(maxv))))):
                    cList.append(deepcopy(a))
            j = k-1
            done2 = False
            while done2 == False:
                if digCnt[j] != digit[j]:
                    done2 = True
                    break
                a[j] = label[j][0]
                digCnt[j] = 0
                j -= 1
                if j < 0:
                    done2 = True
                    break
            if j < 0:
                done = True
                break
            digCnt[j] += 1
            a[j] = label[j][digCnt[j]]
            
    else:
        cList = []
        crange = range(0,size+1)
        aranges = []
        for i in range(nspecies):
            aranges.append(crange)

        p_ranges = product(*aranges)
        for p in p_ranges:
            if sum(p) == size:
                # if not any([list(c) in cList for c in it.permutations(p)]) == True:
                cList.append(list(p))
        cList=cList[1:-1]
                    
    return(cList)

def _does_mapping_exist(v,this_type,atom_pos,atomType,eps):
    """Checks to see if a mapping exists between the vector v and the
      position of any of the atoms of type "this_type".  If a mapping
      exists, then the logical "mapped" is returned .true., otherwise
      .false.

      :args v: Integer array of the position to check mapping for
      :args this_type: Integer that indicates which type of atom that is
            being checked.
      :args atom_pos: 2D integer array of the positions of the basis
            atoms.
      :args atomType: Array of integers of the types of atoms in the
            basis.
      :args eps: Epsilon for checking equivalence.
    """
    mapped = False
    for i in range(len(atomType)):
        if atomType[i] == this_type:
            # if the coordinates are the same, 
            # their difference will be zero for every component
            this_position = atom_pos[i]
            if(numpy.allclose(numpy.array(v), numpy.array(this_position), rtol=eps,atol=eps)):
                mapped = True
                break

    return mapped


def _get_transformations(par_lat):

    """This routine generates the matrices for converting vectors from
      lattice coordinates to cartesion coordinates and vice versa.

      :args par_lat: A 2D integer array that contains the parent lattice vectors
    """

    prim_to_cart = deepcopy(par_lat)
    cart_to_prim = numpy.linalg.inv(numpy.array(prim_to_cart)).tolist()

    return(prim_to_cart,cart_to_prim)

def bring_into_cell(vec,cart_to_latt,latt_to_cart,eps):
    """This subroutine translates a point back into the unit cell in
      lattice coordinates, the coefficients of the point must all be
      less than one and at least zero.
    
      :args vec: A 1D integer array of the atom position vector
      :args cart_to_latt: The matrix that transforms from cartesian to lattice coordinates
      :args latt_to_cart: The matrix that transforms form lattice to cartesian coordinates
      :args eps: Finite precision tolerance.
    """

    # Put the representation of the point into lattice coordinates
    vec = numpy.matmul(numpy.array(cart_to_latt),numpy.array(vec)).tolist()

    # counter to catch compiler bug
    c = 0
    maxc = max(math.ceil(abs(max(vec))),math.ceil(abs(min(vec))))*2

    # If a component >= 1, translate by subtracting a lattice vector
    # If a component < 0, translate by adding a lattice vector
    while any([i > 1.0-eps for i in vec]) or any([i < 0.0-eps for i in vec]):
        c = c +1
        for i in range(len(vec)):
            if vec[i] >= 1-eps:
                vec[i] -= 1
            elif vec[i] < 0-eps:
                vec[i] += 1
        if (c>maxc):
            print("ERROR: loop does not end in bring_into_cell. Probably compiler bug.")
            exit()

    
    # Put the point back into cartesion coordinate representation
    vec = numpy.matmul(numpy.array(latt_to_cart), numpy.array(vec)).tolist()

    return(vec)

def _get_lattice_pointGroup(aVecs, eps=1E-10):
    """This routine returns only the point group of the rather than the
      space group of the given crystal structure.

      :args aVecs: The 2D integer array that contains the parent lattice vectors
      :args eps: (Optional) Finite precision tolerance
    """

    inverse_aVecs = numpy.linalg.inv(numpy.array(aVecs)).tolist()
    # Store the norms of the three lattice vectors
    norm_avecs = []
    for i in range(3):
        norm_avecs.append(numpy.linalg.norm(aVecs[i]).tolist())

    # Decide how many lattice points to look in each direction to get all the 
    # points in a sphere that contains all of the longest _primitive_ vectors
    cell_volume = abs(numpy.dot(aVecs[0],numpy.cross(aVecs[1],aVecs[2])))
    max_norm = max([numpy.linalg.norm(i) for i in aVecs])
    n1 = math.ceil(max_norm*numpy.linalg.norm(numpy.cross(aVecs[1],aVecs[2]))/cell_volume+eps)
    n2 = math.ceil(max_norm*numpy.linalg.norm(numpy.cross(aVecs[2],aVecs[0]))/cell_volume+eps)
    n3 = math.ceil(max_norm*numpy.linalg.norm(numpy.cross(aVecs[0],aVecs[1]))/cell_volume+eps)

    Rvecs = []
    Rlengths = []

    # Store the R vectors that lie within the sphere
    num_Rs = 0
    for i in range(-int(round(n1)), int(round(n1))+1):
        for j in range(-int(round(n2)), int(round(n2))+1):
            for k in range(-int(round(n3)), int(round(n3))+1):
                this_vector = [i*z[0] + j*z[1] + k*z[2] for z in aVecs]
                length = numpy.linalg.norm(this_vector)
                if (length > max_norm + eps):
                    continue # This vector is outside sphere
                num_Rs += 1
                Rvecs.append(this_vector)
                Rlengths.append(length)

    # Try all R vector triplets in the sphere and see which ones are valid 
    # rotations of the original basis vectors.
    # 
    # The length of all vectors must be preserved under a unitary
    # transformation so skip any trial vectors that aren't the same
    # length as the original. We also skip any set of vectors that
    # have the right lengths but do not form a parallelpiped that has
    # the same volume as the original set. Also, note that the we skip
    # sets of vectors that contain the same vector more than once
    # (i.e. the indices i, j, k must be unique).
    num_ops = 0
    lattpg_op = []
    for i in range(num_Rs):
        if (abs(Rlengths[i] - norm_avecs[0]) > eps):
            continue
        for j in range(num_Rs):
            if (abs(Rlengths[j] - norm_avecs[1]) > eps):
                continue
            if (j == i):
                continue
            for k in range(num_Rs):
                if (abs(Rlengths[k] - norm_avecs[2]) > eps):
                    continue
             
                if (k == i or k == j):
                    continue
                if (abs(cell_volume - abs(numpy.linalg.det([Rvecs[i],Rvecs[j],Rvecs[k]]))) > eps):
                    continue
                # Form the new set of "rotated" basis vectors
                new_vectors = [Rvecs[i],Rvecs[j],Rvecs[k]]
                # If the transformation matrix that takes the original set to the new set is
                # an orthogonal matrix then this rotation is a point symmetry of the lattice.
                rotation_matrix = numpy.matmul(inverse_aVecs,new_vectors).tolist()
                # Check orthogonality of rotation matrix by [R][R]^T = [1]
                test_matrix = numpy.matmul(rotation_matrix,numpy.transpose(rotation_matrix)).tolist()
                if (numpy.allclose(test_matrix, [[1,0,0],[0,1,0],[0,0,1]], rtol=eps,atol=eps)): # Found valid rotation
                    num_ops +=  1 # Count number of rotations
                    lattpg_op.append(rotation_matrix)

    return(lattpg_op)
    
def get_spaceGroup(par_lat,atomType,bas_vecs,eps=1E-10,lattcoords = False):

    """This routine takes a crystal structure (basis vectors and basis
      atoms and positions) and returns the point operators and
      fractional translations of the space group. The routine assumes
      that the given crystal structure is already primitive.

      :args par_lat: A 2D integere array that contains the parent lattice vectors
      :args atomType: Integer array representing the type of each basis atom
      :args bas_vecs: A 2D integere array that contains the basis vectors for the cell
      :args eps: (Optional) Finite precisions tolerance
      :args lattcoords: (Optional) True if vectors are in lattice
            coordinates rather than cartesian
    """
    
    # Get number of atoms in the basis
    nAtoms = len(atomType)

    # save original atomic input positions
    atom_pos = deepcopy(bas_vecs)

    # A vector can be represented as either a set of cartesian coordi-
    # nates or as a linear combination of primitive lattice vectors
    # Get transformation matrices to take us back and forth
    (latt_to_cart,cart_to_latt) = _get_transformations(par_lat)
    
    
    # If we're in lattice coordinates Convert the position of the
    # basis atoms from lattice coordinates.
    if lattcoords:
        for i in range(nAtoms):
            atom_pos[i] = numpy.matmul(numpy.array(latt_to_cart),numpy.array(atom_pos[i])).tolist()

    # bring all the basis atoms into the unit cell
    for i in range(len(bas_vecs)):
        bas_vecs[i] = bring_into_cell(bas_vecs[i],cart_to_latt,latt_to_cart,eps)
    # Now find the point group
    lattpg_op = _get_lattice_pointGroup(par_lat)

    # **** Find the elements of the space group ****
    # Count the elements
    sgop_count = 0
    sg_ops = []
    sg_fracts = []
    # Apply each of the point operators in the point group to the crystal
    for iop in range(len(lattpg_op)):
        # rotate atom 1 and store its position in the vector v
        v = numpy.matmul(numpy.array(lattpg_op[iop]),numpy.array(atom_pos[0])).tolist()
        # Loop over all possible fractional translations
        for jAtom in range(nAtoms):
            if (atomType[jAtom] != atomType[0]):
                continue
            fract = [atom_pos[jAtom][i] - v[i] for i in range(3)]
            fract = bring_into_cell(fract, cart_to_latt, latt_to_cart, eps)
            # Is each atom of every type mapped by this rotation + translation?
            for kAtom in range(nAtoms):
                this_type = atomType[kAtom]
                # Rotate and translate each atom        
                v2 = numpy.matmul(numpy.array(lattpg_op[iop]),numpy.array(atom_pos[kAtom])).tolist()
                v2 = [v2[i] + fract[i] for i in range(3)]
                v2 = bring_into_cell(v2, cart_to_latt, latt_to_cart, eps)
                # Try to map this rotated atom onto another the same type
                mapped = _does_mapping_exist(v2, this_type, atom_pos, atomType, eps)
                if not mapped:
                    break # no mapping for this atom

            # if all atoms mapped successfully then count this
            # element (rotation + translation)
            if mapped:
                sgop_count += 1 # Count the number of elements
                sg_fracts.append(fract) # Save the translational part
                sg_ops.append(lattpg_op[iop]) # Store the rotational part
                # loop over fractional translations and try next op
                # By removing the preceding exit, we include fractional translations
                # for non-primitive lattices. (GLWH 10/26/2009)
    return(sg_ops,sg_fracts)
